clear; close all; clc

rng(5);             % seed for reproducible "randomness"
%% 4a)
N = 30;
x = linspace(0,5,N)';
y = 3 * x + 4;              % true output
y = y + randn(N,1);         % add noise

% visualize
figure(1); clf; hold on
plot(x, y, 'ko', 'DisplayName','measurements no outliers')
legend('location', 'eastoutside')

%% 4b)
J = [x, ones(N,1)];
ab = (J' * J) \ (J' * y);                               % solve linear system
plot(x,ab(1) * x+ab(2), 'b', 'DisplayName','L2 norm (no outliers)')

%% 4c)
% introduce outliers
y2      = y;
y2(1)   = 25;
y2(5)   = 30;
y2(end) = 10;
plot(x, y2, 'kx', 'DisplayName','measurements with outliers')

ab2 = (J'*J)\(J'*y2);                     % solve
plot(x, ab2(1) * x + ab2(2), 'r', 'DisplayName','L2 norm (outliers)')

%% 4d)
import casadi.*
opti = Opti();                  % new Opti instance
ab_ca = opti.variable(2);       % decision variable
resid = J * ab_ca - y;          % residuals
opti.minimize(resid' * resid)   % least squares objective
opti.solver('ipopt')
sol = opti.solve();
ab_ca_sol = sol.value(ab_ca);

plot(x, ab_ca_sol(1) * x + ab_ca_sol(2), 'g--', 'DisplayName','casadi L2 (no outliers)')

%% TASK 5
% 5b)

f = [ 0; 0; ones(N,1) ];

A = [    J,     -eye(N) ; ...
        -J,     -eye(N) ; ... 
    zeros(N,2), -eye(N)]; % or move to bounds

b  = [ y ; -y ; zeros( N, 1 ) ];   % without outliers
b2 = [ y2 ; -y2 ; zeros( N, 1 ) ]; % with outliers

ab3 = linprog( f, A, b );
ab4 = linprog( f, A, b2 );

plot(x, ab3(1) * x + ab3(2), 'm', 'DisplayName','L1 norm (no outliers)')
plot(x, ab4(1) * x + ab4(2), 'c', 'DisplayName','L1 norm (outliers)')

%% 5c)
opti = Opti();
ab_ca2 = opti.variable(2);       % parameters
s = opti.variable(N);           % slacks

opti.minimize(sum1(s));
opti.subject_to( -s <= J * ab_ca2 - y <=s)
opti.subject_to(s >= 0)
opti.solver('ipopt')
sol = opti.solve();
ab_ca2_sol = sol.value(ab_ca2);

plot(x, ab_ca2_sol(1) * x + ab_ca2_sol(2), 'y--', 'DisplayName','casadi L1 (no outliers)')

